# Copyright CNRS/Inria/UniCA
# Contributor(s): Eric Debreuve (since 2023)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
import logging as lggg
import typing as h

from logger_36.catalog.config.console_rich import (
    ACTUAL_COLOR,
    DATE_TIME_COLOR,
    ELAPSED_TIME_COLOR,
    EXPECTED_COLOR,
    GRAY_STYLE,
    LEVEL_COLOR,
)
from logger_36.config.message import (
    ACTUAL_PATTERNS,
    ELAPSED_TIME_SEPARATOR,
    EXPECTED_PATTERNS,
    LEVEL_CLOSING,
    WHERE_SEPARATOR,
)
from logger_36.constant.message import TIME_LENGTH
from logger_36.constant.record import SHOW_W_RULE_ATTR
from logger_36.task.format.rule import Rule
from logger_36.type.extension import handler_extension_t
from rich.console import Console as console_t
from rich.console import RenderableType as renderable_t
from rich.markup import escape as EscapedForRich
from rich.text import Text as text_t
from rich.traceback import install as InstallTracebackHandler

_COMMON_TRACEBACK_ARGUMENTS = ("theme", "width")
_EXCLUSIVE_TRACEBACK_ARGUMENTS = (
    "extra_lines",
    "indent_guides",
    "locals_hide_dunder",
    "locals_hide_sunder",
    "locals_max_length",
    "locals_max_string",
    "max_frames" "show_locals",
    "suppress",
    "word_wrap",
)


@dtcl.dataclass(slots=True, repr=False, eq=False)
class console_rich_handler_t(lggg.Handler):
    extension: handler_extension_t = dtcl.field(init=False)
    console: console_t = dtcl.field(init=False)
    FormattedLines: h.Callable[..., tuple[str, str | None]] = dtcl.field(init=False)

    name: dtcl.InitVar[str | None] = None
    level: dtcl.InitVar[int] = lggg.NOTSET
    show_where: dtcl.InitVar[bool] = True
    show_memory_usage: dtcl.InitVar[bool] = False
    message_width: dtcl.InitVar[int] = -1
    formatter: dtcl.InitVar[lggg.Formatter | None] = None
    should_install_traceback: dtcl.InitVar[bool] = False

    rich_kwargs: dtcl.InitVar[dict[str, h.Any] | None] = None

    def __post_init__(
        self,
        name: str | None,
        level: int,
        show_where: bool,
        show_memory_usage: bool,
        message_width: int,
        formatter: lggg.Formatter | None,
        should_install_traceback: bool,
        rich_kwargs: dict[str, h.Any] | None,
    ) -> None:
        """"""
        lggg.Handler.__init__(self)

        self.extension = handler_extension_t(
            name=name,
            show_where=show_where,
            show_memory_usage=show_memory_usage,
            handler=self,
            level=level,
            message_width=message_width,
            formatter=formatter,
        )

        if rich_kwargs is None:
            rich_console_kwargs = {}
        else:
            rich_console_kwargs = rich_kwargs
        rich_traceback_kwargs = {}
        if should_install_traceback:
            for key in rich_console_kwargs:
                if key in _COMMON_TRACEBACK_ARGUMENTS:
                    rich_traceback_kwargs[key] = rich_console_kwargs[key]
                elif key in _EXCLUSIVE_TRACEBACK_ARGUMENTS:
                    rich_traceback_kwargs[key] = rich_console_kwargs[key]
                    del rich_console_kwargs[key]

        self.console = console_t(
            highlight=False,
            force_terminal=True,
            record=True,
            **rich_console_kwargs,
        )
        if should_install_traceback:
            rich_traceback_kwargs["console"] = self.console
            InstallTracebackHandler(**rich_traceback_kwargs)

        self.FormattedLines = self.extension.FormattedLines

    def emit(self, record: lggg.LogRecord, /) -> None:
        """"""
        cls = self.__class__
        if hasattr(record, SHOW_W_RULE_ATTR):
            richer = Rule(record.msg, DATE_TIME_COLOR)
        else:
            first, next_s = self.FormattedLines(record, PreProcessed=EscapedForRich)
            richer = cls.HighlightedVersion(first, next_s, record.levelno)
        self.console.print(richer, crop=False, overflow="ignore")

    def ShowMessage(self, message: str, /) -> None:
        """"""
        self.console.print(message, crop=False, overflow="ignore")

    @classmethod
    def HighlightedVersion(
        cls, first_line: str, next_lines: str | None, log_level: int, /
    ) -> renderable_t:
        """"""
        output = text_t(first_line)

        # Used instead of _CONTEXT_LENGTH which might include \t, thus creating a
        # mismatch between character length and length when displayed in console.
        context_end = first_line.find(LEVEL_CLOSING)
        elapsed_time_separator = first_line.rfind(ELAPSED_TIME_SEPARATOR)
        where_separator = first_line.rfind(
            WHERE_SEPARATOR, context_end, elapsed_time_separator
        )

        output.stylize(DATE_TIME_COLOR, end=TIME_LENGTH)
        output.stylize(
            LEVEL_COLOR[log_level],
            start=TIME_LENGTH,
            end=context_end + 1,
        )
        output.stylize(GRAY_STYLE, start=where_separator, end=elapsed_time_separator)
        output.stylize(ELAPSED_TIME_COLOR, start=elapsed_time_separator)

        if next_lines is not None:
            output.append(next_lines)

        _ = output.highlight_regex(ACTUAL_PATTERNS, style=ACTUAL_COLOR)
        _ = output.highlight_regex(EXPECTED_PATTERNS, style=EXPECTED_COLOR)

        return output



# Copyright CNRS/Inria/UniCA
# Contributor(s): Eric Debreuve (since 2023)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import typing as h

from logger_36.config.message import (
    ELAPSED_TIME_FORMAT,
    LEVEL_CLOSING,
    LEVEL_OPENING,
    MEMORY_FORMAT,
    MESSAGE_MARKER,
)
from logger_36.constant.generic import NOT_PASSED
from logger_36.constant.message import EXPECTED_OP, expected_op_h


def MessageFormat(with_where: bool, with_memory_usage: bool, /) -> str:
    """"""
    output = [
        f"%(asctime)s"
        f"{LEVEL_OPENING}%(level_first_letter)s{LEVEL_CLOSING}\t"
        f"{MESSAGE_MARKER}%(message)s"
    ]

    if with_where:
        output.append("%(where)s")
    output.append(ELAPSED_TIME_FORMAT)
    if with_memory_usage:
        output.append(MEMORY_FORMAT)

    return "".join(output)


def FormattedMessage(
    message: str,
    /,
    *,
    actual: h.Any = NOT_PASSED,
    expected: h.Any | None = None,
    expected_op: expected_op_h = "=",
    with_final_dot: bool = True,
) -> str:
    """"""
    if expected_op not in EXPECTED_OP:
        raise ValueError(
            FormattedMessage(
                'Invalid "expected" section operator',
                actual=expected_op,
                expected=f"One of {str(EXPECTED_OP)[1:-1]}",
            )
        )

    if actual is NOT_PASSED:
        if with_final_dot:
            if message[-1] != ".":
                message += "."
        elif message[-1] == ".":
            message = message[:-1]

        return message

    if message[-1] == ".":
        message = message[:-1]
    actual = _FormattedValue(actual)
    expected = _FormattedValue(expected)

    if with_final_dot:
        dot = "."
    else:
        dot = ""
    return f"{message}: Actual={actual}; Expected{expected_op}{expected}{dot}"


def _FormattedValue(value: h.Any, /, *, should_format_str: bool = True) -> str:
    """"""
    if value is None:
        return "None"

    if isinstance(value, str):
        if should_format_str:
            return f'"{value}"'
        return value

    return str(value)



# Copyright CNRS/Inria/UniCA
# Contributor(s): Eric Debreuve (since 2023)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
import logging as lggg
import sys as sstm
import typing as h
from os import sep as FOLDER_SEPARATOR
from pathlib import Path as path_t

from logger_36.config.message import TIME_FORMAT, WHERE_FORMAT
from logger_36.constant.error import MEMORY_MEASURE_ERROR
from logger_36.constant.handler import HANDLER_CODES
from logger_36.constant.message import NEXT_LINE_PROLOGUE
from logger_36.constant.record import HIDE_WHERE_ATTR, SHOW_WHERE_ATTR
from logger_36.task.format.message import FormattedMessage, MessageFormat
from logger_36.task.measure.chronos import TimeStamp
from logger_36.task.measure.memory import CanCheckMemory

_MEMORY_MEASURE_ERROR = MEMORY_MEASURE_ERROR


@dtcl.dataclass(slots=True, repr=False, eq=False)
class handler_extension_t:
    name: str | None = None
    show_where: bool = True
    show_memory_usage: bool = False
    message_width: int = -1
    FormattedRecord: h.Callable[[lggg.LogRecord], str] = dtcl.field(init=False)

    handler: dtcl.InitVar[lggg.Handler | None] = None
    level: dtcl.InitVar[int] = lggg.NOTSET
    formatter: dtcl.InitVar[lggg.Formatter | None] = None

    def __post_init__(
        self, handler: lggg.Handler | None, level: int, formatter: lggg.Formatter | None
    ) -> None:
        """"""
        global _MEMORY_MEASURE_ERROR

        if self.name in HANDLER_CODES:
            raise ValueError(
                FormattedMessage(
                    "Invalid handler name",
                    actual=self.name,
                    expected=f"a name not in {str(HANDLER_CODES)[1:-1]}",
                )
            )

        if self.name is None:
            self.name = TimeStamp()

        if self.show_memory_usage and not CanCheckMemory():
            self.show_memory_usage = False
            if _MEMORY_MEASURE_ERROR is not None:
                print(_MEMORY_MEASURE_ERROR, file=sstm.stderr)
                _MEMORY_MEASURE_ERROR = None

        handler.setLevel(level)

        if 0 < self.message_width < 5:
            self.message_width = 5
        if formatter is None:
            message_format = MessageFormat(self.show_where, self.show_memory_usage)
            formatter = lggg.Formatter(fmt=message_format, datefmt=TIME_FORMAT)
        handler.setFormatter(formatter)
        self.FormattedRecord = handler.formatter.format

    def FormattedLines(
        self,
        record: lggg.LogRecord,
        /,
        *,
        PreProcessed: h.Callable[[str], str] | None = None,
        should_join_lines: bool = False,
    ) -> tuple[str, str | None]:
        """
        See logger_36.catalog.handler.README.txt.
        """
        record.level_first_letter = record.levelname[0]

        message = record.msg
        if not isinstance(message, str):
            message = str(message)
        original_message = message

        if PreProcessed is not None:
            message = PreProcessed(message)
        if (has_newlines := ("\n" in message)) or (
            (self.message_width > 0) and (message.__len__() > self.message_width)
        ):
            if has_newlines:
                lines = message.splitlines()
                if self.message_width > 0:
                    lines = _WrappedLines(lines, self.message_width)
            else:
                lines = _WrappedLines([message], self.message_width)
            next_lines = NEXT_LINE_PROLOGUE.join(lines[1:])
            next_lines = f"{NEXT_LINE_PROLOGUE}{next_lines}"
            message = lines[0]
        else:
            next_lines = None
        if self.message_width > 0:
            n_missing_s = self.message_width - message.__len__()
            if n_missing_s > 3:
                message += " " + (n_missing_s - 1) * "."
            elif n_missing_s > 0:
                message += n_missing_s * " "

        record.msg = message
        if self.show_where and not hasattr(record, SHOW_WHERE_ATTR):
            hide_where = getattr(record, HIDE_WHERE_ATTR, False)
            if hide_where:
                record.where = ""
            else:
                module = path_t(record.pathname)
                path_was_found = False
                for path in sstm.path:
                    if module.is_relative_to(path):
                        module = module.relative_to(path)
                        path_was_found = True
                        break
                if path_was_found:
                    module = str(module.parent / module.stem)
                    module = module.replace(FOLDER_SEPARATOR, ".")
                else:
                    module = record.module
                record.where = WHERE_FORMAT.format(
                    module=module, funcName=record.funcName, lineno=record.lineno
                )
        first_line = self.FormattedRecord(record).replace("\t", " ")

        # Revert the record message to its original value for subsequent handlers.
        record.msg = original_message

        if should_join_lines:
            if next_lines is None:
                return first_line, None
            else:
                return f"{first_line}{next_lines}", None
        else:
            return first_line, next_lines


def _WrappedLines(lines: list[str], message_width: int, /) -> list[str]:
    """"""
    output = []

    for line in lines:
        while line.__len__() > message_width:
            if all(
                _elm != " " for _elm in line[(message_width - 1) : (message_width + 1)]
            ):
                if line[message_width - 2] == " ":
                    piece, line = (
                        line[: (message_width - 2)].rstrip(),
                        line[(message_width - 1) :],
                    )
                else:
                    piece, line = (
                        line[: (message_width - 1)] + "-",
                        line[(message_width - 1) :],
                    )
            else:
                piece, line = (
                    line[:message_width].rstrip(),
                    line[message_width:].lstrip(),
                )
            output.append(piece)

        output.append(line)

    return output



# Copyright CNRS/Inria/UniCA
# Contributor(s): Eric Debreuve (since 2023)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
import logging as lggg
import sys as sstm
import traceback as tbck
import types as t
import typing as h
from datetime import datetime as dttm
from pathlib import Path as path_t
from traceback import TracebackException as traceback_t

from logger_36.config.issue import ISSUE_CONTEXT_END, ISSUE_CONTEXT_SEPARATOR
from logger_36.config.message import DATE_FORMAT
from logger_36.constant.generic import NOT_PASSED
from logger_36.constant.issue import ORDER, order_h
from logger_36.constant.logger import (
    HIDE_WHERE_KWARG,
    LOGGER_NAME,
    WARNING_LOGGER_NAME,
    WARNING_TYPE_COMPILED_PATTERN,
    logger_handle_h,
)
from logger_36.constant.memory import UNKNOWN_MEMORY_USAGE
from logger_36.constant.record import SHOW_MEMORY_ATTR, SHOW_W_RULE_ATTR
from logger_36.task.format.memory import (
    FormattedUsageWithAutoUnit as FormattedMemoryUsage,
)
from logger_36.task.format.message import FormattedMessage
from logger_36.task.measure.chronos import ElapsedTime
from logger_36.task.measure.memory import CurrentUsage as CurrentMemoryUsage
from logger_36.type.issue import NewIssue, issue_t


@dtcl.dataclass(slots=True, repr=False, eq=False)
class logger_t(lggg.Logger):
    name: dtcl.InitVar[str] = LOGGER_NAME
    level: dtcl.InitVar[int] = lggg.NOTSET
    activate_wrn_interceptions: dtcl.InitVar[bool] = True
    exit_on_error: bool = False
    # Must not be False until at least one handler has been added.
    should_hold_messages: bool = True

    on_hold: list[lggg.LogRecord] = dtcl.field(init=False, default_factory=list)
    last_message_date: str = dtcl.field(init=False, default="")
    any_handler_shows_memory: bool = dtcl.field(init=False, default=False)
    memory_usages: list[tuple[str, int]] = dtcl.field(init=False, default_factory=list)
    context_levels: list[str] = dtcl.field(init=False, default_factory=list)
    staged_issues: list[issue_t] = dtcl.field(init=False, default_factory=list)
    intercepted_wrn_handle: logger_handle_h | None = dtcl.field(
        init=False, default=None
    )
    intercepted_log_handles: dict[str, logger_handle_h] = dtcl.field(
        init=False, default_factory=dict
    )

    def __post_init__(
        self, name: str, level: int, activate_wrn_interceptions: bool
    ) -> None:
        """"""
        lggg.Logger.__init__(self, name)
        self.setLevel(level)
        self.propagate = False

        if activate_wrn_interceptions:
            self._ActivateWarningInterceptions()

    def _ActivateWarningInterceptions(self) -> None:
        """
        The log message will not appear if called from __post_init__ since there are no
        handlers yet.
        """
        if self.intercepted_wrn_handle is None:
            logger = lggg.getLogger(WARNING_LOGGER_NAME)
            self.intercepted_wrn_handle = logger.handle
            logger.handle = t.MethodType(_HandleForWarnings(self), logger)

            lggg.captureWarnings(True)
            self.info("Warning Interception: ON", **HIDE_WHERE_KWARG)

    def _DeactivateWarningInterceptions(self) -> None:
        """"""
        if self.intercepted_wrn_handle is not None:
            logger = lggg.getLogger(WARNING_LOGGER_NAME)
            logger.handle = self.intercepted_wrn_handle
            self.intercepted_wrn_handle = None

            lggg.captureWarnings(False)
            self.info("Warning Interception: OFF", **HIDE_WHERE_KWARG)

    def ToggleWarningInterceptions(self, state: bool, /) -> None:
        """"""
        if state:
            self._ActivateWarningInterceptions()
        else:
            self._DeactivateWarningInterceptions()

    def ToggleLogInterceptions(self, state: bool, /) -> None:
        """"""
        if state:
            self.ToggleLogInterceptions(False)

            all_loggers = [lggg.getLogger()] + [
                lggg.getLogger(_nme)
                for _nme in self.manager.loggerDict
                if _nme not in (self.name, WARNING_LOGGER_NAME)
            ]
            for logger in all_loggers:
                self.intercepted_log_handles[logger.name] = logger.handle
                logger.handle = t.MethodType(
                    _HandleForInterceptions(logger, self), logger
                )

            intercepted = sorted(self.intercepted_log_handles.keys())
            if intercepted.__len__() > 0:
                as_str = ", ".join(intercepted)
                self.info(
                    f"Now Intercepting LOGs from: {as_str}",
                    **HIDE_WHERE_KWARG,
                )
        elif self.intercepted_log_handles.__len__() > 0:
            for name, handle in self.intercepted_log_handles.items():
                logger = lggg.getLogger(name)
                logger.handle = handle
            self.intercepted_log_handles.clear()
            self.info("Log Interception: OFF", **HIDE_WHERE_KWARG)

    @property
    def max_memory_usage(self) -> int:
        """"""
        if self.memory_usages.__len__() > 0:
            return max(tuple(zip(*self.memory_usages))[1])
        return UNKNOWN_MEMORY_USAGE

    @property
    def max_memory_usage_full(self) -> tuple[str, int]:
        """"""
        if self.memory_usages.__len__() > 0:
            where_s, usages = zip(*self.memory_usages)
            max_usage = max(usages)

            return where_s[usages.index(max_usage)], max_usage

        return "?", UNKNOWN_MEMORY_USAGE

    def AddHandler(self, handler: lggg.Handler, should_hold_messages: bool, /) -> None:
        """"""
        self.should_hold_messages = should_hold_messages
        lggg.Logger.addHandler(self, handler)

        extension = getattr(handler, "extension", None)
        if extension is None:
            show_memory_usage = False
        else:
            show_memory_usage = getattr(extension, SHOW_MEMORY_ATTR, False)
        if show_memory_usage:
            self.any_handler_shows_memory = True

        extension = getattr(handler, "extension", handler.name)
        if isinstance(extension, str):
            name = extension
        else:
            name = getattr(extension, "name", "Anonymous")
        self.info(
            f'New handler "{name}" with class "{type(handler).__name__}" and '
            f"level {lggg.getLevelName(handler.level)}",
            **HIDE_WHERE_KWARG,
        )

    def handle(self, record: lggg.LogRecord, /) -> None:
        """"""
        if (not self.should_hold_messages) and (self.on_hold.__len__() > 0):
            for hold in self.on_hold:
                lggg.Logger.handle(self, hold)
            self.on_hold.clear()

        record.elapsed_time = ElapsedTime()

        if self.any_handler_shows_memory or not self.hasHandlers():
            # Memory usage is also added if there are no handlers yet, just in case.
            usage = CurrentMemoryUsage()
            self.memory_usages.append(
                (f"{record.module}.{record.funcName}.{record.lineno}", usage)
            )

            value, unit = FormattedMemoryUsage(usage, 1)
            record.memory_usage = f"{value}{unit}"

        date = dttm.now().strftime(DATE_FORMAT)
        if date != self.last_message_date:
            self.last_message_date = date
            # levelno: Added for management by logging.Logger.handle.
            date_record = lggg.makeLogRecord(
                {
                    "name": self.name,
                    "levelno": lggg.INFO,
                    "msg": f"DATE: {date}",
                    SHOW_W_RULE_ATTR: True,
                }
            )
            if self.should_hold_messages:
                self.on_hold.append(date_record)
            else:
                lggg.Logger.handle(self, date_record)

        if self.should_hold_messages:
            self.on_hold.append(record)
        else:
            lggg.Logger.handle(self, record)

        if self.exit_on_error and (record.levelno in (lggg.ERROR, lggg.CRITICAL)):
            sstm.exit(1)

    def AddContextLevel(self, new_level: str, /) -> None:
        """"""
        self.context_levels.append(new_level)

    def AddedContextLevel(self, new_level: str, /) -> logger_t:
        """
        Meant to be used as:
        with self.AddedContextLevel("new level"):
            ...
        """
        self.AddContextLevel(new_level)
        return self

    def StageIssue(
        self,
        message: str,
        /,
        *,
        actual: h.Any = NOT_PASSED,
        expected: h.Any | None = None,
    ) -> None:
        """"""
        context = ISSUE_CONTEXT_SEPARATOR.join(self.context_levels)
        issue = NewIssue(
            context, ISSUE_CONTEXT_END, message, actual=actual, expected=expected
        )
        self.staged_issues.append(issue)

    @property
    def has_staged_issues(self) -> bool:
        """"""
        return self.staged_issues.__len__() > 0

    def CommitIssues(
        self,
        level: int | type[Exception],
        /,
        *,
        order: order_h = "when",
    ) -> None:
        """"""
        if not self.has_staged_issues:
            return

        if order not in ORDER:
            raise ValueError(
                FormattedMessage(
                    "Invalid commit order",
                    actual=order,
                    expected=f"One of {str(ORDER)[1:-1]}",
                )
            )

        if order == "when":
            issues = self.staged_issues
        else:  # order == "context"
            issues = sorted(self.staged_issues, key=lambda _elm: _elm.context)
        issues = "\n".join(issues)

        if isinstance(level, type) and issubclass(level, Exception):
            try:
                raise level("\n" + issues)
            except Exception as exception:
                lines = ["Traceback (most recent call last):"] + tbck.format_stack()[
                    :-1
                ]
                lines[-1] = lines[-1][:-1]
                lines.extend(tbck.format_exception_only(exception))
                print("\n".join(lines), file=sstm.stderr)
                sstm.exit(1)

        self.log(level, issues, stacklevel=2)
        self.staged_issues.clear()

    def ShowMessage(self, message: str, /) -> None:
        """"""
        for handler in self.handlers:
            ShowMessage = getattr(handler, "ShowMessage", None)
            if ShowMessage is not None:
                ShowMessage(message)

    def __enter__(self) -> None:
        """"""
        pass

    def __exit__(
        self,
        exc_type: Exception | None,
        exc_value: str | None,
        traceback: traceback_t | None,
        /,
    ) -> bool:
        """"""
        _ = self.context_levels.pop()
        return False


def _HandleForWarnings(interceptor: lggg.Logger, /) -> logger_handle_h:
    """"""

    def handle_p(_: lggg.Logger, record: lggg.LogRecord, /) -> None:
        pieces = WARNING_TYPE_COMPILED_PATTERN.match(record.msg)
        if pieces is None:
            # The warning message does not follow the default format.
            interceptor.handle(record)
            return

        GetPiece = pieces.group
        path = GetPiece(1)
        line = GetPiece(2)
        kind = GetPiece(3)
        message = GetPiece(4)

        duplicate = lggg.makeLogRecord(record.__dict__)
        duplicate.msg = f"{kind}: {message}"
        duplicate.pathname = path
        duplicate.module = path_t(path).stem
        duplicate.funcName = "?"
        duplicate.lineno = line

        interceptor.handle(duplicate)

    return handle_p


def _HandleForInterceptions(
    intercepted: lggg.Logger, interceptor: lggg.Logger, /
) -> logger_handle_h:
    """"""

    def handle_p(_: lggg.Logger, record: lggg.LogRecord, /) -> None:
        duplicate = lggg.makeLogRecord(record.__dict__)
        duplicate.msg = f"{record.msg} :{intercepted.name}:"
        interceptor.handle(duplicate)

    return handle_p



# Copyright CNRS/Inria/UniCA
# Contributor(s): Eric Debreuve (since 2023)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from logger_36.config.memory import LENGTH_100, MAX_N_SAMPLES
from logger_36.constant.logger import HIDE_WHERE_KWARG
from logger_36.constant.memory import storage_units_h
from logger_36.instance import LOGGER
from logger_36.task.format.memory import FormattedUsage, UsageBar
from logger_36.task.format.message import FormattedMessage


def LogMemoryUsages(
    *,
    unit: storage_units_h | None = "a",
    decimals: int = None,
    max_n_samples: int | None = MAX_N_SAMPLES,
    length_100: int = LENGTH_100,
) -> None:
    """"""
    if not LOGGER.any_handler_shows_memory:
        return

    where_s, usages = zip(*LOGGER.memory_usages)

    where, max_usage = LOGGER.max_memory_usage_full
    value, unit = FormattedUsage(max_usage, unit=unit, decimals=decimals)
    title = f"Memory Usage: Max={value}{unit} near {where}\n"

    if isinstance(max_n_samples, int):
        if max_n_samples < 1:
            raise ValueError(
                FormattedMessage(
                    "Invalid maximum number of samples",
                    actual=max_n_samples,
                    expected=1,
                    expected_op=">=",
                )
            )

        where_s = list(where_s)
        usages = list(usages)
        while usages.__len__() > max_n_samples:
            index = usages.index(min(usages))
            del where_s[index]
            del usages[index]

    usages = tuple(round(_elm, 1) for _elm in usages)
    max_usage = max(usages)

    plot = []
    max_where_length = max(map(len, where_s))
    usages_as_str = tuple(map(lambda _elm: f"{_elm:_}", usages))
    max_usage_length = max(map(len, usages_as_str))
    for where, usage, usage_as_str in zip(where_s, usages, usages_as_str):
        bar = UsageBar(usage, max_usage, length_100=length_100)
        plot.append(
            f"{where:{max_where_length}} "
            f"{bar:{length_100}} "
            f"{usage_as_str: >{max_usage_length}}"
        )
    plot = "\n".join(plot)

    LOGGER.info(title + plot, **HIDE_WHERE_KWARG)


def LogMaximumMemoryUsage(
    *, unit: storage_units_h | None = "a", decimals: int = None
) -> None:
    """
    unit: b or None=bytes, k=kilo, m=mega, g=giga, a=auto
    """
    if LOGGER.any_handler_shows_memory:
        where, max_usage = LOGGER.max_memory_usage_full
        value, unit = FormattedUsage(max_usage, unit=unit, decimals=decimals)
        LOGGER.info(
            f"Max. Memory Usage: {value}{unit} near {where}", **HIDE_WHERE_KWARG
        )



# Copyright CNRS/Inria/UniCA
# Contributor(s): Eric Debreuve (since 2023)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
import logging as lggg
import typing as h
from pathlib import Path as path_t

from logger_36.constant.record import SHOW_W_RULE_ATTR
from logger_36.task.format.rule import RuleAsText
from logger_36.type.extension import handler_extension_t


@dtcl.dataclass(slots=True, repr=False, eq=False)
class file_handler_t(lggg.FileHandler):

    extension: handler_extension_t = dtcl.field(init=False)
    FormattedLines: h.Callable[..., tuple[str, str | None]] = dtcl.field(init=False)

    name: dtcl.InitVar[str | None] = None
    level: dtcl.InitVar[int] = lggg.NOTSET
    show_where: dtcl.InitVar[bool] = True
    show_memory_usage: dtcl.InitVar[bool] = False
    message_width: dtcl.InitVar[int] = -1
    formatter: dtcl.InitVar[lggg.Formatter | None] = None

    path: dtcl.InitVar[path_t | None] = None
    handler_args: dtcl.InitVar[tuple[h.Any, ...] | None] = None
    handler_kwargs: dtcl.InitVar[dict[str, h.Any] | None] = None

    def __post_init__(
        self,
        name: str | None,
        level: int,
        show_where: bool,
        show_memory_usage: bool,
        message_width: int,
        formatter: lggg.Formatter | None,
        path: path_t | None,
        handler_args: tuple[h.Any],
        handler_kwargs: dict[str, h.Any] | None,
    ) -> None:
        """"""
        lggg.FileHandler.__init__(self, path, *handler_args, **handler_kwargs)

        self.extension = handler_extension_t(
            name=name,
            show_where=show_where,
            show_memory_usage=show_memory_usage,
            handler=self,
            level=level,
            message_width=message_width,
            formatter=formatter,
        )

        self.FormattedLines = self.extension.FormattedLines

    def emit(self, record: lggg.LogRecord, /) -> None:
        """"""
        if hasattr(record, SHOW_W_RULE_ATTR):
            message = RuleAsText(record.msg)
        else:
            message, _ = self.FormattedLines(record, should_join_lines=True)
        print(message, file=self.stream)
        self.stream.flush()

    def ShowMessage(self, message: str, /) -> None:
        """"""
        print(message, file=self.stream)
        self.stream.flush()



# Copyright CNRS/Inria/UniCA
# Contributor(s): Eric Debreuve (since 2023)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

GPU_LOGGING_ERROR = (
    "GPU details cannot be logged because the Tensorflow and/or Tensorrt package(s) "
    "(https://www.tensorflow.org/, https://developer.nvidia.com/tensorrt)"
    "is/are not installed or not importable."
)

MEMORY_MEASURE_ERROR = (
    "Memory usage cannot be shown because the Psutil package "
    "(https://psutil.readthedocs.io/en/latest/)"
    "is not installed or not importable."
)

MISSING_RICH_ERROR = (
    "The Rich console handler is not available because the Rich package "
    "(https://rich.readthedocs.io/en/stable/) "
    "is not installed or not importable. "
    "Falling back to the raw console."
)



# Copyright CNRS/Inria/UniCA
# Contributor(s): Eric Debreuve (since 2023)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
import logging as lggg
import typing as h

from logger_36.catalog.config.console_rich import DATE_TIME_COLOR

try:
    from rich.console import Console as console_t
    from rich.console import ConsoleOptions as console_options_t
    from rich.markup import escape as EscapedForRich
    from rich.terminal_theme import DEFAULT_TERMINAL_THEME
except ModuleNotFoundError:
    console_t = EscapedForRich = DEFAULT_TERMINAL_THEME = None

from logger_36.catalog.handler.console_rich import console_rich_handler_t
from logger_36.constant.record import SHOW_W_RULE_ATTR
from logger_36.task.format.rule import Rule, RuleAsText
from logger_36.type.extension import handler_extension_t


class can_show_message_p(h.Protocol):
    def ShowMessage(self, message: str, /) -> None: ...


interface_h = can_show_message_p | h.Callable[[str], None]


@dtcl.dataclass(slots=True, repr=False, eq=False)
class generic_handler_t(lggg.Handler):

    extension: handler_extension_t = dtcl.field(init=False)
    console: console_t | None = None
    console_options: console_options_t | None = None
    FormattedLines: h.Callable[..., tuple[str, str | None]] = dtcl.field(init=False)
    ShowMessage: h.Callable[[str], None] = lambda _arg: None

    name: dtcl.InitVar[str | None] = None
    level: dtcl.InitVar[int] = lggg.NOTSET
    show_where: dtcl.InitVar[bool] = True
    show_memory_usage: dtcl.InitVar[bool] = False
    message_width: dtcl.InitVar[int] = -1
    formatter: dtcl.InitVar[lggg.Formatter | None] = None

    supports_html: dtcl.InitVar[bool] = False
    rich_kwargs: dtcl.InitVar[dict[str, h.Any] | None] = None
    interface: dtcl.InitVar[interface_h | None] = None  # Cannot be None actually.

    def __post_init__(
        self,
        name: str | None,
        level: int,
        show_where: bool,
        show_memory_usage: bool,
        message_width: int,
        formatter: lggg.Formatter | None,
        supports_html: bool,
        rich_kwargs: dict[str, h.Any] | None,
        interface: interface_h | None,
    ) -> None:
        """"""
        lggg.Handler.__init__(self)

        self.extension = handler_extension_t(
            name=name,
            show_where=show_where,
            show_memory_usage=show_memory_usage,
            handler=self,
            level=level,
            message_width=message_width,
            formatter=formatter,
        )

        if supports_html and (console_t is not None):
            if rich_kwargs is None:
                rich_kwargs = {}
            self.console = console_t(
                highlight=False,
                force_terminal=True,
                **rich_kwargs,
            )
            self.console_options = self.console.options.update(
                overflow="ignore", no_wrap=True
            )

        self.FormattedLines = self.extension.FormattedLines

        self.ShowMessage = getattr(
            interface, can_show_message_p.ShowMessage.__name__, interface
        )

    def emit(self, record: lggg.LogRecord, /) -> None:
        """"""
        if self.console is None:
            if hasattr(record, SHOW_W_RULE_ATTR):
                message = RuleAsText(record.msg)
            else:
                message, _ = self.FormattedLines(record, should_join_lines=True)
        else:
            if hasattr(record, SHOW_W_RULE_ATTR):
                richer = Rule(record.msg, DATE_TIME_COLOR)
            else:
                first, next_s = self.FormattedLines(record, PreProcessed=EscapedForRich)
                richer = console_rich_handler_t.HighlightedVersion(
                    first, next_s, record.levelno
                )
            segments = self.console.render(richer, options=self.console_options)

            # Inspired from the code of: rich.console.export_html.
            html_segments = []
            for text, style, _ in segments:
                if text == "\n":
                    html_segments.append("\n")
                else:
                    if style is not None:
                        style = style.get_html_style(DEFAULT_TERMINAL_THEME)
                        if (style is not None) and (style.__len__() > 0):
                            text = f'<span style="{style}">{text}</span>'
                    html_segments.append(text)
            if html_segments[-1] == "\n":
                html_segments = html_segments[:-1]

            # /!\ For some reason, the widget splits the message into lines, place each
            # line inside a pre tag, and set margin-bottom of the first and list lines
            # to 12px. This can be seen by printing self.contents.toHtml(). To avoid the
            # unwanted extra margins, margin-bottom is set to 0 below.
            message = (
                "<pre style='margin-bottom:0px'>" + "".join(html_segments) + "</pre>"
            )

        self.ShowMessage(message)



# Copyright CNRS/Inria/UniCA
# Contributor(s): Eric Debreuve (since 2023)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
import logging as lggg
import re as regx
import typing as h
from html.parser import HTMLParser as html_parser_t
from pathlib import Path as path_t

try:
    from rich.console import Console as console_t
except ModuleNotFoundError:
    console_t = None

from logger_36.instance import LOGGER

_BODY_END_PATTERN = r"</[bB][oO][dD][yY]>(.|\n)*$"


@dtcl.dataclass(slots=True, repr=False, eq=False)
class html_reader_t(html_parser_t):
    source: str = ""
    inside_body: bool = dtcl.field(init=False, default=False)
    body_position_start: tuple[int, int] = dtcl.field(init=False, default=(-1, -1))
    body_position_end: tuple[int, int] = dtcl.field(init=False, default=(-1, -1))
    pieces: list[str] = dtcl.field(init=False, default_factory=list)

    def __post_init__(self) -> None:
        """"""
        html_parser_t.__init__(self)
        self.source = self.source.strip()
        self.feed(self.source)

    def handle_starttag(self, tag: str, attrs: list[tuple[str, str | None]], /) -> None:
        """"""
        if tag == "body":
            self.body_position_start = self.getpos()
            self.inside_body = True

    def handle_endtag(self, tag: str, /) -> None:
        """"""
        if tag == "body":
            self.body_position_end = self.getpos()
            self.inside_body = False

    def handle_data(self, data: str, /) -> None:
        """"""
        if self.inside_body:
            self.pieces.append(data)

    @property
    def body(self) -> str:
        """"""
        output = self.source.splitlines()
        output = "\n".join(
            output[self.body_position_start[0] : (self.body_position_end[0] + 1)]
        )
        output = output[self.body_position_start[1] :]
        output = regx.sub(_BODY_END_PATTERN, "", output, count=1)

        return output.strip()

    @property
    def body_as_text(self) -> str:
        """"""
        return "".join(self.pieces).strip()


def SaveLOGasHTML(path: str | path_t | h.TextIO = None) -> None:
    """
    From first console handler found.
    """
    cannot_save = "Cannot save logging record as HTML"

    if console_t is None:
        LOGGER.warning(f"{cannot_save}: The Rich console cannot be imported.")
        return

    if path is None:
        for handler in LOGGER.handlers:
            if isinstance(handler, lggg.FileHandler):
                path = path_t(handler.baseFilename).with_suffix(".htm")
                break
        if path is None:
            LOGGER.warning(f"{cannot_save}: No file handler to build a filename from.")
            return
        if path.exists():
            LOGGER.warning(
                f'{cannot_save}: Automatically generated path "{path}" already exists.'
            )
            return
    elif isinstance(path, str):
        path = path_t(path)

    actual_file = isinstance(path, path_t)
    if actual_file and path.exists():
        LOGGER.warning(f'{cannot_save}: File "{path}" already exists.')
        return

    console = None
    found = False
    for handler in LOGGER.handlers:
        console = getattr(handler, "console", None)
        if found := isinstance(console, console_t):
            break

    if found:
        html = console.export_html()
        if actual_file:
            with open(path, "w") as accessor:
                accessor.write(html)
        else:
            path.write(html)
    else:
        LOGGER.warning(f"{cannot_save}: No handler has a RICH console.")



# Copyright CNRS/Inria/UniCA
# Contributor(s): Eric Debreuve (since 2023)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import logging as lggg
import sys as sstm
from pathlib import Path as path_t

from logger_36.catalog.handler.console import console_handler_t
from logger_36.catalog.handler.file import file_handler_t
from logger_36.catalog.handler.generic import generic_handler_t, interface_h
from logger_36.constant.error import MISSING_RICH_ERROR
from logger_36.constant.handler import HANDLER_CODES, handler_codes_h
from logger_36.instance import LOGGER
from logger_36.task.format.message import FormattedMessage

try:
    from logger_36.catalog.handler.console_rich import console_rich_handler_t

    _MISSING_RICH_ERROR = None
except ModuleNotFoundError:
    from logger_36.catalog.handler.console import (
        console_handler_t as console_rich_handler_t,
    )

    _MISSING_RICH_ERROR = MISSING_RICH_ERROR


def AddGenericHandler(
    interface: interface_h,
    /,
    *,
    logger: lggg.Logger | None = None,
    name: str | None = None,
    level: int = lggg.INFO,
    show_where: bool = True,
    show_memory_usage: bool = False,
    message_width: int = -1,
    formatter: lggg.Formatter | None = None,
    supports_html: bool = False,
    should_hold_messages: bool = False,
    **kwargs,
) -> None:
    """"""
    if logger is None:
        logger = LOGGER

    handler = generic_handler_t(
        name=name,
        level=level,
        show_where=show_where,
        show_memory_usage=show_memory_usage,
        message_width=message_width,
        formatter=formatter,
        supports_html=supports_html,
        rich_kwargs=kwargs,
        interface=interface,
    )
    logger.AddHandler(handler, should_hold_messages)


def AddConsoleHandler(
    *,
    logger: lggg.Logger | None = None,
    name: str | None = None,
    level: int = lggg.INFO,
    show_where: bool = True,
    show_memory_usage: bool = False,
    message_width: int = -1,
    formatter: lggg.Formatter | None = None,
    should_hold_messages: bool = False,
) -> None:
    """"""
    if logger is None:
        logger = LOGGER

    handler = console_handler_t(
        name=name,
        level=level,
        show_where=show_where,
        show_memory_usage=show_memory_usage,
        message_width=message_width,
        formatter=formatter,
    )
    logger.AddHandler(handler, should_hold_messages)


def AddRichConsoleHandler(
    *,
    logger: lggg.Logger | None = None,
    name: str | None = None,
    level: int = lggg.INFO,
    show_where: bool = True,
    show_memory_usage: bool = False,
    message_width: int = -1,
    formatter: lggg.Formatter | None = None,
    should_hold_messages: bool = False,
    should_install_traceback: bool = False,
    **kwargs,
) -> None:
    """"""
    global _MISSING_RICH_ERROR
    if _MISSING_RICH_ERROR is not None:
        print(_MISSING_RICH_ERROR, file=sstm.stderr)
        _MISSING_RICH_ERROR = None

    if logger is None:
        logger = LOGGER

    handler = console_rich_handler_t(
        name=name,
        level=level,
        show_where=show_where,
        show_memory_usage=show_memory_usage,
        message_width=message_width,
        formatter=formatter,
        should_install_traceback=should_install_traceback,
        rich_kwargs=kwargs,
    )
    logger.AddHandler(handler, should_hold_messages)


def AddFileHandler(
    path: str | path_t,
    /,
    *args,
    logger: lggg.Logger | None = None,
    name: str | None = None,
    level: int = lggg.INFO,
    show_where: bool = True,
    show_memory_usage: bool = False,
    message_width: int = -1,
    formatter: lggg.Formatter | None = None,
    should_hold_messages: bool = False,
    **kwargs,
) -> None:
    """"""
    if isinstance(path, str):
        path = path_t(path)
    if path.exists():
        raise ValueError(f"File or folder already exists: {path}.")
    if logger is None:
        logger = LOGGER

    handler = file_handler_t(
        name=name,
        level=level,
        show_where=show_where,
        show_memory_usage=show_memory_usage,
        message_width=message_width,
        formatter=formatter,
        path=path,
        handler_args=args,
        handler_kwargs=kwargs,
    )
    logger.AddHandler(handler, should_hold_messages)


def SetLOGLevel(
    level: int,
    /,
    *,
    logger: lggg.Logger | None = None,
    which: handler_codes_h | str = "a",
) -> None:
    """
    which: g=generic, c=console, f=file, a=all, str=name.
    """
    if logger is None:
        logger = LOGGER

    which_is_name = which not in HANDLER_CODES
    found = False
    for handler in logger.handlers:
        if (
            (which == "a")
            or ((which == "g") and isinstance(handler, generic_handler_t))
            or (
                (which == "c")
                and isinstance(handler, (console_handler_t, console_rich_handler_t))
            )
            or ((which == "f") and isinstance(handler, file_handler_t))
            or (which == handler.name)
        ):
            handler.setLevel(level)
            if which_is_name:
                return
            found = True

    if not found:
        raise ValueError(
            FormattedMessage(
                "Handler not found",
                actual=which,
                expected=f"{str(HANDLER_CODES)[1:-1]}, or a handler name",
            )
        )



# Copyright CNRS/Inria/UniCA
# Contributor(s): Eric Debreuve (since 2023)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from logger_36.constant.memory import STORAGE_UNITS, storage_units_h
from logger_36.task.format.message import FormattedMessage

_KILO_UNIT = 1000.0
_MEGA_UNIT = _KILO_UNIT * 1000.0
_GIGA_UNIT = _MEGA_UNIT * 1000.0

_BLOCKS_PARTIAL = ("", "▏", "▎", "▍", "▌", "▋", "▊")
_BLOCK_FULL = "▉"


def FormattedUsage(
    usage: int,
    /,
    *,
    unit: storage_units_h | None = "a",
    decimals: int = None,
) -> tuple[int | float, str]:
    """
    unit: b or None=bytes, k=kilo, m=mega, g=giga, a=auto
    """
    if (unit is None) or (unit == "b"):
        value = usage
        unit = "B"
    elif unit == "k":
        value = _Rounded(usage / _KILO_UNIT, decimals)
        unit = "KB"
    elif unit == "m":
        value = _Rounded(usage / _MEGA_UNIT, decimals)
        unit = "MB"
    elif unit == "g":
        value = _Rounded(usage / _GIGA_UNIT, decimals)
        unit = "GB"
    elif unit == "a":
        value, unit = FormattedUsageWithAutoUnit(usage, decimals)
    else:
        raise ValueError(
            FormattedMessage(
                "Invalid unit", actual=unit, expected=str(STORAGE_UNITS)[1:-1]
            )
        )

    return value, unit


def FormattedUsageWithAutoUnit(
    usage: int, decimals: int | None, /
) -> tuple[int | float, str]:
    """"""
    if usage > _GIGA_UNIT:
        return _Rounded(usage / _GIGA_UNIT, decimals), "GB"

    if usage > _MEGA_UNIT:
        return _Rounded(usage / _MEGA_UNIT, decimals), "MB"

    if usage > _KILO_UNIT:
        return _Rounded(usage / _KILO_UNIT, decimals), "KB"

    return usage, "B"


def UsageBar(
    usage: int | float, max_usage: int | float, /, *, length_100: int = 10
) -> str:
    """"""
    length = (usage / max_usage) * length_100
    n_full_s = int(length)
    return (
        n_full_s * _BLOCK_FULL
        + _BLOCKS_PARTIAL[int((2.0 / 3.0) * int(10 * (length - n_full_s)))]
    )


def _Rounded(value: float, decimals: int | None, /) -> int | float:
    """"""
    if decimals == 0:
        decimals = None

    return round(value, ndigits=decimals)



# Copyright CNRS/Inria/UniCA
# Contributor(s): Eric Debreuve (since 2023)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import re as rgex
from importlib import util
from pathlib import Path as path_t
from typing import Dict

from setuptools import setup

HERE = path_t(__file__).parent.resolve()
LOCAL_DOCUMENTATION = HERE / "documentation" / "wiki"
ASCIIDOC_DESCRIPTION = "description.asciidoc"


def DescriptionFromAsciidoc(documentation: path_t, /) -> Dict[str, str]:
    """"""
    output = {}

    pattern = rgex.compile(r":([A-Z_]+): +(.+)\n?", flags=rgex.ASCII)

    with open(documentation) as accessor:
        for line in accessor.readlines():
            if (match := pattern.fullmatch(line)) is not None:
                name = match.group(1)
                value = match.group(2)
                output[name] = value

    return output


DESCRIPTION = DescriptionFromAsciidoc(LOCAL_DOCUMENTATION / ASCIIDOC_DESCRIPTION)
PACKAGES = [
    DESCRIPTION["IMPORT_NAME"],
    f"{DESCRIPTION['IMPORT_NAME']}.catalog",
    f"{DESCRIPTION['IMPORT_NAME']}.catalog.config",
    f"{DESCRIPTION['IMPORT_NAME']}.catalog.handler",
    f"{DESCRIPTION['IMPORT_NAME']}.catalog.logging",
    f"{DESCRIPTION['IMPORT_NAME']}.config",
    f"{DESCRIPTION['IMPORT_NAME']}.constant",
    f"{DESCRIPTION['IMPORT_NAME']}.task",
    f"{DESCRIPTION['IMPORT_NAME']}.task.format",
    f"{DESCRIPTION['IMPORT_NAME']}.task.measure",
    f"{DESCRIPTION['IMPORT_NAME']}.type",
]
EXCLUDED_PACKAGES = (
    f"{DESCRIPTION['IMPORT_NAME']}.documentation",
)
ENTRY_POINTS = {
    "console_scripts": [],
    "gui_scripts": [],
}


long_description = (HERE / "README.rst").read_text(encoding="utf-8")

repository_url = (
    f"https://"
    f"{DESCRIPTION['REPOSITORY_SITE']}/"
    f"{DESCRIPTION['REPOSITORY_USER']}/"
    f"{DESCRIPTION['REPOSITORY_NAME']}/"
)
documentation_url = f"{repository_url}/{DESCRIPTION['DOCUMENTATION_SITE']}"


def CheckCoherenceBetweenDeclarationAndReality() -> None:
    """"""
    folders = [DESCRIPTION["IMPORT_NAME"]]
    for node in (HERE / DESCRIPTION["IMPORT_NAME"]).rglob("*"):
        if node.is_dir() and not str(node).startswith("."):
            node = node.relative_to(HERE)
            node = ".".join(node.parts)
            if not (
                (node in EXCLUDED_PACKAGES)
                or any(node.startswith(_fld + ".") for _fld in EXCLUDED_PACKAGES)
            ):
                folders.append(node)
    folders = sorted(folders)

    packages = sorted(PACKAGES)
    if packages != folders:
        raise ValueError(
            f"Mismatch between declared and found packages:\n"
            f"    - Declared=\n      {packages}\n"
            f"    - Found=\n      {folders}\n"
            f"    - Undeclared=\n      {set(folders).difference(packages)}\n"
            f"    - Nonexistent=\n      {set(packages).difference(folders)}"
        )


def Version() -> str:
    """"""
    where = HERE / DESCRIPTION["IMPORT_NAME"] / "version.py"
    spec = util.spec_from_file_location(where.stem, where)
    module = spec.loader.load_module(spec.name)

    output = module.__version__
    if isinstance(output, str) and rgex.fullmatch(r"20[0-9]{2}\.[1-9][0-9]*", output):
        return output

    raise ValueError(f"{output}: Invalid version")


def Requirements() -> tuple[str, ...]:
    """"""
    if not (HERE / "requirements.txt").is_file():
        return ()

    with open(HERE / "requirements.txt") as accessor:
        output = accessor.readlines()

    return tuple(output)


if __name__ == "__main__":
    #
    CheckCoherenceBetweenDeclarationAndReality()
    # fmt: off
    setup(
        author=DESCRIPTION["AUTHOR"],
        author_email=DESCRIPTION["EMAIL"],
        #
        name=DESCRIPTION["PYPI_NAME"],
        description=DESCRIPTION["SHORT_DESCRIPTION"],
        long_description=long_description,
        long_description_content_type="text/x-rst",
        license=DESCRIPTION["LICENSE_SHORT"],
        version=Version(),
        #
        classifiers=[
            f"Topic :: {DESCRIPTION['PYPI_TOPIC']}",
            f"Intended Audience :: {DESCRIPTION['PYPI_AUDIENCE']}",
            f"License :: OSI Approved :: {DESCRIPTION['LICENCE_LONG']} ({DESCRIPTION['LICENSE_SHORT']})",
            f"Programming Language :: Python :: {DESCRIPTION['PY_VERSION_MAJOR']}",
            f"Development Status :: {DESCRIPTION['PYPI_STATUS']}",
        ],
        keywords=DESCRIPTION["KEYWORDS"],
        #
        url=repository_url,
        project_urls={
            "Documentation": documentation_url,
            "Source": repository_url,
        },
        #
        packages=PACKAGES,
        entry_points=ENTRY_POINTS,
        python_requires=f">={DESCRIPTION['PY_VERSION_MIN']}",
        install_requires=Requirements(),
    )
